#! /usr/bin/env ruby

# play.rb: Written by Tadayoshi Funaba 1999-2001
# $Id: play.rb,v 1.15 2001-12-03 21:33:01+09 tadf Exp $

require 'smf'
include  SMF

class Devseq < File

  def dumpbuf
    syswrite(@buf)
    @buf = ''
  end

  def putbuf(s)
    @buf ||= ''
    if @buf.size > 4096
      dumpbuf
    end
    @buf << s
  end

  def start_timer() putbuf([0x81, 4, 0, 0, 0].pack('CCCCI')) end
  def stop_timer() putbuf([0x81, 3, 0, 0, 0].pack('CCCCI')) end
  def wait(time) putbuf([0x81, 2, 0, 0, time].pack('CCCCI')) end
  def midiout(dev, byte) putbuf([0x5, byte, dev, 0].pack('CCCC')) end

  case RUBY_PLATFORM
  when /freebsd/no
    SNDCTL_SEQ_CTRLRATE = 0xc0045103
    SNDCTL_SEQ_NRMIDIS  = 0x4004510b
  when /linux/no
    SNDCTL_SEQ_CTRLRATE = 0xc0045103
    SNDCTL_SEQ_NRMIDIS  = 0x8004510b
  else
    fail
  end

  def ctrlrate
    n = [0].pack('i')
    ioctl(SNDCTL_SEQ_CTRLRATE, n)
    n.unpack('i')[0]
  end

  def nrmidis
    n = [0].pack('i')
    ioctl(SNDCTL_SEQ_NRMIDIS, n)
    n.unpack('i')[0]
  end

end

module SMF

  class Sequence

    class Play < XSCallback

      def header(format, ntrks, division)
	@division = division
      end

      def track_start
	@seq = Devseq.open('/dev/sequencer', 'w')
	@nrmidis = @seq.nrmidis
	if @nrmidis.zero?
	  fail 'No MIDI ports'
	end
	@ctrlrate = @seq.ctrlrate
	@tempo = 500000
	@offset = 0
	@cur = 0
	@seq.dumpbuf
	@seq.start_timer
      end

      def track_end
	@seq.dumpbuf
	@seq.stop_timer
	@seq.close
      end

      def delta(delta)
	@offset += delta
	if delta.nonzero?
	  delta2 = delta.to_f *
	    @tempo * @ctrlrate / 1000000 / @division
	  cur = @cur + delta2
	  if cur.to_i > @cur.to_i
	    @seq.wait(cur.to_i)
	  end
	  @cur = cur
	end
      end

      def noteoff(ch, note, vel)
	@seq.midiout(0, ch | 0x80)
	@seq.midiout(0, note)
	@seq.midiout(0, vel)
      end

      def noteon(ch, note, vel)
	@seq.midiout(0, ch | 0x90)
	@seq.midiout(0, note)
	@seq.midiout(0, vel)
      end

      def polyphonickeypressure(ch, note, val)
	@seq.midiout(0, ch | 0xa0)
	@seq.midiout(0, note)
	@seq.midiout(0, val)
      end

      def controlchange(ch, num, val)
	@seq.midiout(0, ch | 0xb0)
	@seq.midiout(0, num)
	@seq.midiout(0, val)
      end

      def programchange(ch, num)
	@seq.midiout(0, ch | 0xc0)
	@seq.midiout(0, num)
      end

      def channelpressure(ch, val)
	@seq.midiout(0, ch | 0xd0)
	@seq.midiout(0, val)
      end

      def pitchbendchange(ch, val)
	@seq.midiout(0, ch | 0xe0)
	lsb =  val       & 0x7f
	msb = (val >> 7) & 0x7f
	@seq.midiout(0, lsb)
	@seq.midiout(0, msb)
      end

      def channelmodemessage(ch, num, val)
	@seq.midiout(0, ch | 0xb0)
	@seq.midiout(0, num)
	@seq.midiout(0, val)
      end

      def exclusivefx(data)
	data.each_byte do |x|
	  @seq.midiout(0, x)
	end
      end

      def exclusivef0(data)
	@seq.midiout(0, 0xf0)
	exclusivefx(data)
      end

      def exclusivef7(data)
	exclusivefx(data)
      end

      def settempo(tempo)
	@tempo = tempo
      end

    end

    def play
      WS.new(self, Play.new).read
    end

  end

end

def usage
  $stderr.puts 'usage: play file'
  exit 1
end

usage unless $*.size == 1
file = $*.shift

sq = Sequence.decodefile(file)
nsq = Sequence.new(0, sq.division)
ntr = Track.new
sq.each do |tr|
  tr.pop
  tr.each do |ev|
    ntr << ev
  end
end
nsq << ntr
nsq.play
