#! /usr/bin/env ruby

# smf2text.rb: Written by Tadayoshi Funaba 1999-2001
# $Id: smf2text.rb,v 1.8 2001-12-03 21:33:01+09 tadf Exp $

require 'smf'
include  SMF

module SMF

  class Sequence

    class EncodeText < XSCallback

      def header(format, ntrks, division)
	@s = format("Sequence %d %d %d\n", format, ntrks, division)
      end

      def track_start() @offset = 0; @s << "Track\n" end

      def delta(delta) @offset += delta end

      def noteoff(ch, note, vel)
	@s << format("%d NoteOff %d %d %d\n", @offset, ch, note, vel)
      end

      def noteon(ch, note, vel)
	@s << format("%d NoteOn %d %d %d\n", @offset, ch, note, vel)
      end

      def polyphonickeypressure(ch, note, val)
	@s << format("%d PolyphonicKeyPressure %d %d %d\n",
		     @offset, ch, note, val)
      end

      def controlchange(ch, num, val)
	@s << format("%d ControlChange %d %d %d\n", @offset, ch, num, val)
      end

      def programchange(ch, num)
	@s << format("%d ProgramChange %d %d\n", @offset, ch, num)
      end

      def channelpressure(ch, val)
	@s << format("%d ChannelPressure %d %d\n", @offset, ch, val)
      end

      def pitchbendchange(ch, val)
	@s << format("%d PitchbendChange %d %d\n", @offset, ch, val)
      end

      def channelmodemessage(ch, num, val)
	@s << format("%d ChannelModeMessage %d %d %d\n", @offset, ch, num, val)
      end

      def exclusive(name, data)
	@s << format('%d %s', @offset, name)
	data.each_byte do |c|
	  @s << format(' %d', c)
	end
	@s << "\n"
      end

      private :exclusive

      def exclusivef0(data) exclusive('ExclusiveF0', data) end
      def exclusivef7(data) exclusive('ExclusiveF7', data) end

      def sequencenumber(num)
	@s << format("%d SequenceNumber %d\n", @offset, num)
      end

      def text(name, text)
	@s << format("%d %s %s\n", @offset, name, text.inspect)
      end

      private :text

      def generalpurposetext(text) text('GeneralPurposeText', text) end
      def copyrightnotice(text) text('CopyrightNotice', text) end
      def trackname(text) text('TrackName', text) end
      def instrumentname(text) text('InstrumentName', text) end
      def lyric(text) text('Lyric', text) end
      def marker(text) text('Marker', text) end
      def cuepoint(text) text('CuePoint', text) end
      def programname(text) text('ProgramName', text) end
      def devicename(text) text('DeviceName', text) end
      def text0a(text) text('Text0A', text) end
      def text0b(text) text('Text0B', text) end
      def text0c(text) text('Text0C', text) end
      def text0d(text) text('Text0D', text) end
      def text0e(text) text('Text0E', text) end
      def text0f(text) text('Text0F', text) end

      def channelprefix(ch)
	@s << format("%d ChannelPrefix %d\n", @offset, ch)
      end

      def midiport(num)
	@s << format("%d MidiPort %d\n", @offset, num)
      end

      def endoftrack
	@s << format("%d EndOfTrack\n", @offset)
      end

      def settempo(tempo)
	@s << format("%d SetTempo %d\n", @offset, tempo)
      end

      def smpteoffset(hr, mn, se, fr, ff)
	@s << format("%d SMPTEOffset %d %d %d %d %d\n",
		     @offset, hr, mn, se, fr, ff)
      end

      def timesignature(nn, dd, cc, bb)
	@s << format("%d TimeSignature %d %d %d %d\n", @offset, nn, dd, cc, bb)
      end

      def keysignature(sf, mi)
	@s << format("%d KeySignature %d %d\n", @offset, sf, mi)
      end

      def sequencerspecific(data)
	@s << format('%d SequencerSpecific', @offset)
	data.each_byte do |c|
	  @s << format(' %d', c)
	end
	@s << "\n"
      end

      def result() @s end

    end

    def encodetext
      WS.new(self, EncodeText.new).read
    end

    def encodetextio(io)
      io.write(encodetext)
    end

    def encodetextfile(fn)
      open(fn, 'w') do |io|
	encodetextio(io)
      end
    end

  end

end

def usage
  $stderr.puts 'usage: smf2text ifile [ofile]'
  exit 1
end

usage unless $*.size >= 1 and $*.size <= 2
ifile = $*.shift
ofile = $*.shift
ofile ||= File.basename(ifile, '.mid') + '.txt'

sq = Sequence.decodefile(ifile)
sq.encodetextfile(ofile)
