#! /usr/bin/env ruby

# smf2wav.rb: Written by Tadayoshi Funaba 1999-2002
# $Id: smf2wav.rb,v 1.13 2002-04-29 22:01:49+09 tadf Exp $

require 'smf'
include  SMF

class Wave

  def initialize(rate=8000)
    @rate = rate
    @wave = []
  end

  def sin(start, stop, freq, amp)
    (start..stop).each do |i|
      r = amp * Math.sin(2 * Math::PI * freq * (i.to_f / @rate))
      @wave[i] ||= 0.0
      @wave[i] += r
    end
  end

  def dump
    out = [ 'RIFF', @wave.length + 36, 'WAVE', 'fmt', 16, 1, 1, @rate, @rate,
      1, 8, 'data', @wave.length].pack('A4 V A4 A4 V v v V V v v A4 V')
    peak = 0
    @wave.each do |x|
      peak = x if x and x > peak
    end
    @wave.each_with_index do |x, i|
      @wave[i] = unless x then 128 else 128 + (127 * x / peak) end
    end
    out << @wave.pack('C*')
    out
  end

end

module SMF

  class Sequence

    class EncodeWav < XSCallback

      FREQ = []
      (0..127).each do |n|
	FREQ << 440 * 2**((n-69.0)/12)
      end

      def initialize
	@tempo = [[0, 500000]]
	@rate = 8000
	@wave = Wave.new(@rate)
      end

      def header(format, ntrks, division)
	@division = division
      end

      def track_start
	@offset = 0
	@noteon = []
      end

      def delta(delta) @offset += delta end

      def noteoff(ch, note, vel)
	return if ch == 9	# GM per.
	if val = @noteon[ch].delete(note)
	  start, vel = val
	  @wave.sin(pos(start, @rate), pos(@offset, @rate), FREQ[note], vel)
	end
      end

      def noteon(ch, note, vel)
	return if ch == 9	# GM per.
	if vel == 0
	  noteoff(ch, note, vel)
	else
	  @noteon[ch] ||= {}
	  @noteon[ch][note] ||= [@offset, vel]
	end
      end

      def settempo(tempo)
	@tempo.unshift([@offset, tempo])
      end

      def pos2(offset, tempo, rate)
	(offset * tempo / @division) * rate / 1000000
      end

      def pos(offset, rate)
	po = 0
	@tempo.each do |of, te|
	  if offset > of
	    po += pos2(offset - of, te, rate)
	    offset = of
	  end
	end
	po
      end

      def result() @wave.dump end

    end

    def encodewav
      WS.new(self, EncodeWav.new).read
    end

    def encodewavio(io)
      io.binmode.write(encodewav)
    end

    def encodewavfile(fn)
      open(fn, 'w') do |io|
	encodewavio(io)
      end
    end

  end

end

def usage
  $stderr.puts 'usage: smf2wav ifile [ofile]'
  exit 1
end

usage unless $*.size >= 1 and $*.size <= 2
ifile = $*.shift
ofile = $*.shift
ofile ||= File.basename(ifile, '.mid') + '.wav'

sq = Sequence.decodefile(ifile)
sq.encodewavfile(ofile)
